local bor = bit.bor
local band = bit.band
local tinsert = table.insert
local tremove= table.remove
local tsort= table.sort

local fmod = math.fmod

MovixLib_Gridfus = {}

function MovixLib_Gridfus.ResetColumnOrder (Headers)
 local n = 0;
 if Headers then
    for i, col in ipairs(Headers) do
        n = n +1
        col.pos = n
    end
 end
 tsort(Headers, mycolsort);
end
MovixLib_Grid_ResetColumnOrder = MovixLib_Gridfus.ResetColumnOrder

function MovixLib_Gridfus.CreateHeadersFromData(source)
    local datasource = MovixLib_GetDataSource(source)
    local DHeaders = MovixLib_GetTable()
    local fnames = MovixLib_GetTable()
    for ind, row in pairs(datasource) do
        local datarow
        if type(row) ~= "table" then
         datarow = {row};
        else
         datarow = row;
        end
        local i = 0
        for fname, fvalue in pairs(datarow) do
            if not fnames[fname] then
                fnames[fname] = true
                i = i + 1;
                local colsize = 25
                local celloptions = MovixLib_NewCellOptions(fname, fvalue)
                MovixLib_SetCellOption(celloptions, "size", colsize)
                MovixLib_SetCellOption(celloptions, "pos", i)
                tinsert(DHeaders, celloptions);
                --table.insert(p.Headers, celloptions);
            end
        end
        i = 0;
    end
    MovixLib_Grid_ResetColumnOrder(DHeaders)
    MovixLib_ReleaseTable(fnames)
    return DHeaders  
end
MovixLib_Grid_CreateHeadersFromData = MovixLib_Gridfus.CreateHeadersFromData

function MovixLib_FillData(grid)
 --grid.source = source;
 local Headers=grid.Headers
 if #Headers == 0 then return end
 local datasource = MovixLib_GetDataSource(Movixlib_GetGridParam(grid.gridid, "DataSource"))
 if grid.data == nil then grid.data = MovixLib_GetTable() end
 if #grid.data > 0 then MovixLib_WipeTable(grid.data,50) end
 
 datasource = MovixLib_DataPrepare(grid, datasource)
 datasource = MovixLib_FilterGridData(datasource, Headers)

 grid.rowcount = #datasource;
 
 if Headers and grid.GridParams then
    if grid.GridParams.GroupCol then
        --datasource = MovixLib_GroupData(datasource, grid)
    end
    if grid.GridParams.sortcol then
        MovixLib_SortColParent = grid
        MovixLib_SortCols = grid.GridParams.sortcol
        tsort(datasource, MovixLib_SortTables);
    end
 end
 grid.data = datasource
end

function MovixLib_GroupData(datasource, grid)
    local GroupCol = grid.GridParams.GroupCol
    if not GroupCol then return datasource end
    local Headers=grid.Headers
    local data = MovixLib_GetTable()
    --local grpcol = MovixLib_NewCellOptions ("grptext", "", "string")
    --MovixLib_SetCellOption(grpcol, "pos", 1)
    --table.insert(Headers, 1, grpcol)
    --local i = 0
    --for coli, col in pairs(grid.Headers) do
        --i = i +1;
        --col.pos = i;
    --end
    
    if not grid.grptbl then grid.grptbl= MovixLib_GetTable() end
    local grptbl = grid.grptbl;
    
    MovixLib_SortColParent = grid
    MovixLib_SortColIndex = GroupCol 
    MovixLib_SortColOrder = true
    tsort(datasource, MovixLib_SortTables);
    local col = Headers[GroupCol]
    local act_grpval = MovixLib_GetTransformedValue(datasource[1][col.val], col.valfoo, col.type) or "NIL"
    local grpcolval
    for ind, row in ipairs(datasource) do
        grpcolval = MovixLib_GetTransformedValue(row[col.val], col.valfoo, col.type) or "NIL"
        
        if not grptbl[grpcolval] then 
            grptbl[grpcolval] = MovixLib_GetTable() 
        end
        grptbl[grpcolval].count = (grptbl[grpcolval].count or 0) + 1
        if act_grpval ~= grpcolval then
         tinsert(  data,
                        ind,
                        {
                            MovixLibID = "GRP"..act_grpval ,
                            [col.val] = act_grpval .." ("..grptbl[act_grpval ].count..")",
                        }
                     )
           act_grpval = grpcolval 
        end
        --if grptbl[grpcolval].close then
          --  table.remove(datasource, ind)
        --end
    end
        if not grptbl[grpcolval] then 
            grptbl[grpcolval] = MovixLib_GetTable() 
        end

        act_grpval = grpcolval 
         tinsert(  data,
                        {
                            MovixLibID = "GRP"..act_grpval ,
                            [col.val] = act_grpval .." ("..grptbl[act_grpval ].count..")",
                        }
                     )
    
    return data
end

function MovixLib_DataPrepare(p, datasource)
 local data = MovixLib_GetTable()
 local hasselection = nil
 if type (datasource) ~= "table" then
  datasource = {datasource}
 end
 for ind, row in pairs (datasource) do
    local MovixLibSelected = nil
    if type(row) ~= "table" then -- ROW IS NOT A TABLE BUT ONLY ONE VALUE
     if p.SelectedRow and p.SelectedRow[ind] then
        hasselection = true
        MovixLibSelected = true
     else
        MovixLibSelected = false
     end
     tinsert(data, {MovixLibID = ind, Field = ind, Value = row, MovixLibSelected = MovixLibSelected})
    
    else -- ROW IS A TABLE
    
     local newrow = MovixLib_TblCpy(row)
     newrow.MovixLibID = ind;
     for i, col in ipairs (p.Headers) do -- WALK THROUGH COLUMNS
        local colval = nil;
        if type (row[col.val]) == "table" then
            --newrow[col.val] = "#TBL#"
        elseif type (row[col.val]) == "function" then
            newrow[col.val] = "#FOO#"
        end
     end
     if p.SelectedRow and p.SelectedRow[ind] then
        newrow.MovixLibSelected = true
        hasselection = true
     else
        newrow.MovixLibSelected = false
     end
     tinsert(data, newrow)
    end
 end
 if not hasselection and p.Headers[1] and p.Headers[1].val == "MovixLibSelected" then 
        tremove(p.Headers, 1)
        MovixLib_Grid_ResetColumnOrder(p.Headers)
 end
 return data
end


function MovixLib_FilterGridData(datasource, Headers)
 local data = MovixLib_GetTable()
 for ind, row in pairs (datasource) do
    local showline = true
    if Headers then -- IF HEADER ALLREADY SET
        for i, col in ipairs (Headers) do -- WALK THROUGH COLUMNS
             -- IF FILTER DEFINED FOR THAT COLUMN ONLY ADD ROW IF NOT FILTERED
            if col.filtertbl then
                local filtertext = MovixLib_GetFilterText(col, row[col.val])
                if col.filtertbl[filtertext] ~= nil then
                    if col.filtertbl[filtertext] == false then
                        showline = false
                    end
                end
            end
        end
    end
    
    if showline ~= false then
        tinsert(data, row)
    end
 end
 --MovixLib_WipeTable(datasource)
 return data
end


function MovixLib_GridMenu_OnParentSizeChanged(self)
  if self.old_OnSizeChanged then
   self.old_OnSizeChanged(self)
  end
  self:SetWidth(self:GetRight()-self:GetLeft())
  self:SetHeight(self:GetTop()-self:GetBottom())
  MovixLib_RefreshGrid(self, true, false)
end

function MovixLib_GridMenu_OnParentShowScript(self)
  if self.old_OnShow then
   self.old_OnShow(self)
  end
  MovixLib_FillData(self);
  MovixLib_RefreshGrid(self, true)
end

function MovixLib_GridMenu_OnParentHideScript(self)
end

function MovixLib_GridMenu_OnParentHideScripta(self)
  if self.old_OnHide then
   self.old_OnHide(self)
  end
  local p = self
  if p.lines then 
    for k, gl in ipairs(p.lines) do 
        if gl.cols then
            for ind, ctrl in ipairs(gl.cols) do
                local col = gl.cols[ind]
                MovixLib_ReleaseControlObject(ctrl:GetName())
                col.ComputedW = nil
                col.ComputedH = nil
                col.Lpoint = nil
                col.Rpoint = nil
                col:Hide()
                gl.cols[ind] = nil
            end
            gl.lineoptions = nil
            gl.parent = nil
            gl.cols = nil
            --gl.OddEven = nil
            gl:Hide()
            p.lines[k] = nil
        end
    end
    p.lines= nil
  end
end

--[[
         MovixLib_DrawHeaders(p)
         
Function to create Grid Headerline
]]-- 


function MovixLib_DrawHeaders(p, rebuild, resetfilters)
   local scrollchild, hslider, vslider, hsliderTexture, vsliderTexture , bg, h, hl, hlc, cv, cvf, srt;
   local pwidth = p:GetWidth();
   local pheight = p:GetHeight();
   local i = 0;
   local x = 0;
   local n = 0;
   local totalwidth = 0;
   local datarow;
   p.rebuild = rebuild;
   p:EnableMouse(true);   
   p:SetScript("OnMouseUp", function(self, button) if button == "RightButton" then MovixLib_GridMenu(self, button) end end);
   local SizeChangeScript = p:GetScript("OnSizeChanged");
   if SizeChangeScript ~= MovixLib_GridMenu_OnParentSizeChanged then
      p.old_OnSizeChanged = p:GetScript("OnSizeChanged");
      p:SetScript("OnSizeChanged", MovixLib_GridMenu_OnParentSizeChanged);
   end
   local ShowScript = p:GetScript("OnShow");
   if ShowScript ~= MovixLib_GridMenu_OnParentShowScript then
      p.old_OnShow = p:GetScript("OnShow");
      p:SetScript("OnShow", MovixLib_GridMenu_OnParentShowScript);
   end
   local HideScript = p:GetScript("OnHide");
   if HideScript  ~= MovixLib_GridMenu_OnParentHideScript then
      p.old_OnHide = p:GetScript("OnHide");
      p:SetScript("OnHide", MovixLib_GridMenu_OnParentHideScript);
   end
   
   local headerdef = p.Headers;
   tsort(headerdef, mycolsort);
   
   if p.HeaderLine == nil then -- FIRST CALL -> CREATE
    p.HeaderLine = getglobal (p:GetName().."HeaderLine");
    if p.HeaderLine == nil then
     p.HeaderLine = CreateFrame("ScrollFrame", p:GetName().."Scroll", p, "MovixLib_ScrollFrame_Template");
     --- HEADER SEPARATOR LINE
     hlc = CreateFrame("FRAME", p:GetName().."LC", p.HeaderLine);
     local hl = nil;
     hl = hlc:CreateTexture(p:GetName().."TRC"); -- TOP RIGHT CORNER
     hl:SetDrawLayer("BACKGROUND");
     hl = hlc:CreateTexture(p:GetName().."BRC"); -- BOTTOM RIGHT CORNER
     hl:SetDrawLayer("BACKGROUND");
     hl = hlc:CreateTexture(p:GetName().."HHL");  -- Header SEPARATOR LINE
     hl:SetDrawLayer("BACKGROUND");
     hl = hlc:CreateTexture(p:GetName().."HHLS"); -- Header SEPARATOR LINE SHADOW
     hl:SetDrawLayer("BACKGROUND");
     hl = hlc:CreateTexture(p:GetName().."HBL");  -- Bottom SEPARATOR LINE
     hl:SetDrawLayer("BACKGROUND");
     hl = hlc:CreateTexture(p:GetName().."HBLS"); -- Bottom SEPARATOR LINE SHADOW
     hl:SetDrawLayer("BACKGROUND");
     hl = hlc:CreateTexture(p:GetName().."VRL");  -- Bottom SEPARATOR LINE
     hl:SetDrawLayer("BACKGROUND");
     hl = hlc:CreateTexture(p:GetName().."VRLS"); -- Bottom SEPARATOR LINE SHADOW
     hl:SetDrawLayer("BACKGROUND");
     
     hlc:SetScript("OnMouseDown", MovixLib_StartResizeHeader);
     hlc:SetScript("OnMouseUp", MovixLib_StopResizeHeader);
     hlc:SetScript("OnEnter", MovixLib_OnEnterColSep);
     hlc:SetScript("OnLeave", MovixLib_OnLeaveColSep);
     hlc:RegisterForDrag("LeftButton");
     hlc:SetMovable(true);
     hlc:EnableMouse(true);
     hlc.parent = p;
    end
   end
   
   local LEdgeInset = Movixlib_GetGridParam(p.gridid, "LEdgeInset")
   local REdgeInset = Movixlib_GetGridParam(p.gridid, "REdgeInset") 
   local TEdgeInset = Movixlib_GetGridParam(p.gridid, "TEdgeInset")
   local BEdgeInset = Movixlib_GetGridParam(p.gridid, "BEdgeInset")
   local SliderSize = Movixlib_GetGridParam(p.gridid, "SliderSize")
   local headerheight = Movixlib_GetGridParam(p.gridid, "headerheight")
   local LineHeight = Movixlib_GetGridParam(p.gridid, "LineHeight")
   local ColSepWidth = Movixlib_GetGridParam(p.gridid, "ColSepWidth")
   local LabelFontSize = Movixlib_GetGridParam(p.gridid, "LabelFontSize")
   local GridBGcolor = Movixlib_GetGridParam(p.gridid, "GridBGcolor")
   local HeaderBGcolor = Movixlib_GetGridParam(p.gridid, "HeaderBGcolor")
   local OddLineColor = Movixlib_GetGridParam(p.gridid, "OddLineColor")
   local ShadowColor1 = Movixlib_GetGridParam(p.gridid, "ShadowColor1")
   local LightenColor1 = Movixlib_GetGridParam(p.gridid, "LightenColor1")
   local ShadowColor2 = Movixlib_GetGridParam(p.gridid, "ShadowColor2")
   local LightenColor2 = Movixlib_GetGridParam(p.gridid, "LightenColor2")
   local HeaderTextColor = Movixlib_GetGridParam(p.gridid, "HeaderTextColor")

   local h = p.HeaderLine;
   scrollchild = getglobal(h:GetName().."Scrollchild");
   hslider = getglobal(h:GetName().."_HSlider");
   vslider = getglobal(h:GetName().."_VSlider");
   hsliderTexture = getglobal(hslider:GetName().."ThumbTexture")
   local scrollbarleftarrow = getglobal(hslider:GetName().."ScrollLeftButton")
   local scrollbarrightarrow = getglobal(hslider:GetName().."ScrollRightButton")

   hlc = getglobal(p:GetName().."LC"); -- HEADERLINE DELIMITER
   bg = getglobal(h:GetName().."HeaderBG");
   local gbg = getglobal(h:GetName().."GridBG");
   
   h:SetPoint("TOPLEFT", LEdgeInset, -TEdgeInset);
   h:SetWidth(pwidth-(LEdgeInset+REdgeInset+SliderSize));
   h:SetHeight(pheight-(TEdgeInset+BEdgeInset+SliderSize));
   
   hslider:ClearAllPoints();
   hslider:SetPoint("TOPLEFT", h, "BOTTOMLEFT", SliderSize ,0);
   hslider:SetWidth(pwidth-(LEdgeInset+REdgeInset)-SliderSize-3-(2*SliderSize));
   hslider:SetHeight(SliderSize);
   hsliderTexture:SetWidth(SliderSize);
   hsliderTexture:SetHeight(SliderSize);
   hsliderTexture:SetTexture("interface\\AddOns\\GridLib\\gfxbigslider");
   hsliderTexture:SetTexCoord(0.2,0.65,0.15,0.7)
   scrollbarrightarrow:SetNormalTexture("interface\\AddOns\\GridLib\\gfx\\arrow-r")
   scrollbarleftarrow:SetNormalTexture("interface\\AddOns\\GridLib\\gfx\\arrow-l")
   scrollbarrightarrow:SetDisabledTexture("interface\\AddOns\\GridLib\\gfx\\arrow-r-d")
   scrollbarleftarrow:SetDisabledTexture("interface\\AddOns\\GridLib\\gfx\\arrow-l-d")
   scrollbarrightarrow:SetHighlightTexture("interface\\AddOns\\GridLib\\gfx\\arrow-r-h")
   scrollbarleftarrow:SetHighlightTexture("interface\\AddOns\\GridLib\\gfx\\arrow-l-h")
   scrollbarrightarrow:SetPushedTexture("interface\\AddOns\\GridLib\\gfx\\arrow-r-p")
   scrollbarleftarrow:SetPushedTexture("interface\\AddOns\\GridLib\\gfx\\arrow-l-p")
   scrollbarrightarrow:SetWidth(SliderSize);
   scrollbarrightarrow:SetHeight(SliderSize);
   scrollbarleftarrow:SetWidth(SliderSize);
   scrollbarleftarrow:SetHeight(SliderSize);



   
   vslider:Hide();

   scrollchild:SetHeight(h:GetHeight()-3);
   
   bg:ClearAllPoints();
   bg:SetPoint("TOPLEFT", 0, 0);
   bg:SetHeight(headerheight);
   bg:SetWidth(h:GetWidth()-2);
   bg:SetTexture("interface\\AddOns\\GridLib\\gfx\\titleBar")
   bg:SetVertexColor(unpack(HeaderBGcolor));

   
   
   gbg:ClearAllPoints();
   gbg:SetPoint("TOPLEFT", 0, -headerheight-3);
   gbg:SetHeight(h:GetHeight()-headerheight-6);
   gbg:SetWidth(h:GetWidth()-2);
   gbg:SetTexture(unpack(GridBGcolor));
 
   hlc:SetPoint("TOPLEFT", 0, -headerheight);
   hlc:SetWidth(h:GetWidth()+SliderSize);
   hlc:SetHeight(3);
   
   local hl = getglobal(p:GetName().."BRC");  -- BOTTOM RIGHT CORNER
   hl:SetPoint("BOTTOMRIGHT", 0, -pheight+headerheight+TEdgeInset+BEdgeInset+3);
   hl:SetWidth(SliderSize);
   hl:SetHeight(SliderSize);
   hl:SetTexture(unpack(ShadowColor1));
   local hl = getglobal(p:GetName().."TRC");  -- TOP RIGHT CORNER
   hl:SetPoint("BOTTOMRIGHT", 0, 3);
   hl:SetWidth(SliderSize);
   hl:SetHeight(headerheight);
   hl:SetTexture(unpack(ShadowColor1));
   local hl = getglobal(p:GetName().."VRL");
   hl:SetPoint("TOPRIGHT", -SliderSize, headerheight);
   hl:SetWidth(1);
   hl:SetHeight(h:GetHeight()+SliderSize);
   hl:SetTexture(unpack(LightenColor1));
   local hl = getglobal(p:GetName().."VRLS");
   hl:SetPoint("TOPRIGHT", -1-SliderSize, headerheight);
   hl:SetWidth(2);
   hl:SetHeight(h:GetHeight()+SliderSize);
   hl:SetTexture(unpack(ShadowColor1));
   local hl = getglobal(p:GetName().."HHL");
   hl:SetPoint("TOPLEFT", 0, 0);
   hl:SetWidth(hlc:GetWidth());
   hl:SetHeight(1);
   hl:SetTexture(unpack(LightenColor1));
   local hl = getglobal(p:GetName().."HHLS");
   hl:SetPoint("TOPLEFT", 0, -2);
   hl:SetWidth(hlc:GetWidth());
   hl:SetHeight(2);
   hl:SetTexture(unpack(ShadowColor1));
   local hl = getglobal(p:GetName().."HBL");
   hl:SetPoint("BOTTOMLEFT", 0, -pheight+SliderSize+headerheight+TEdgeInset+BEdgeInset+3);
   hl:SetWidth(hlc:GetWidth());
   hl:SetHeight(1);
   hl:SetTexture(unpack(LightenColor1));
   local hl = getglobal(p:GetName().."HBLS");
   hl:SetPoint("BOTTOMLEFT", 0, -pheight+SliderSize+headerheight+TEdgeInset+BEdgeInset+4);
   hl:SetWidth(hlc:GetWidth());
   hl:SetHeight(2);
   hl:SetTexture(unpack(ShadowColor1));
   
   ---- DRAW COLUMN HEADERS
   for ind, col in ipairs(headerdef) do
    --if col.visible then
      i = i + 1;
      local ht = getglobal(h:GetName()..i.."LBL");
      if ht == nil then 
        ht = CreateFrame("BUTTON", h:GetName()..i.."LBL", scrollchild, "MovixLib_ColmunHeaderTemplate");
        cvf = CreateFrame("FRAME", h:GetName()..i.."VS", scrollchild);
        cv = cvf:CreateTexture(h:GetName()..i.."VT");
        cv:SetDrawLayer("BACKGROUND");
        cv = cvf:CreateTexture(h:GetName()..i.."VTS");
        cv:SetDrawLayer("BACKGROUND");
        
        srt = ht:CreateTexture(h:GetName()..i.."SORT");
        srt:SetDrawLayer("BACKGROUND");
        srt:SetAlpha(0.3)
        
        srt = ht:CreateFontString(ht:GetName().."sorttext", "OVERLAY", "GameFontHighlightSmall");
        srt:SetAlpha(0.8)
        
        ht:RegisterForDrag("LeftButton");
        ht:RegisterForClicks("AnyUp");
        ht:SetMovable(true);
        ht:EnableMouse(true);
    
        ht:SetScript("OnMouseUp", MovixLib_ClickHeader);
        ht:SetScript("OnEnter", MovixLib_OnEnterGridHeader)
        ht:SetScript("OnLeave", MovixLib_OnLeaveGridHeader)
        ht:SetScript("OnDragStart", MovixLib_StartColDrag);
        ht:SetScript("OnDragStop", MovixLib_StopColDrag);
        ht:SetScript("OnReceiveDrag", MovixLib_OnReceiveDrag);

        cvf = getglobal(h:GetName()..i.."VS"); 
        cvf:SetMovable(true);
        cvf:EnableMouse(true);
        cvf:SetScript("OnMouseDown", MovixLib_StartResizeCol);
        cvf:SetScript("OnMouseUp", MovixLib_StopResizeCol);
        cvf:SetScript("OnEnter", MovixLib_OnEnterColSep);
        cvf:SetScript("OnLeave", MovixLib_OnLeaveColSep);
       
        cvf:RegisterForDrag("LeftButton");
      end
      ht:Show();
      
      local srti = getglobal(h:GetName()..i.."SORT");
      local srtt = getglobal(ht:GetName().."sorttext")
      srti:ClearAllPoints()
      srtt:ClearAllPoints()
      if col.justify == "RIGHT" then
            srti:SetPoint("TOPLEFT", -1, 0)
            srtt:SetPoint("TOPLEFT", srti, "TOPLEFT", -1, -4)
      else
            srti:SetPoint("TOPRIGHT", 1, 0)
            srtt:SetPoint("TOPRIGHT", srti, "TOPRIGHT", 1, -4)
      end
      srti:SetWidth(7)
      srti:SetHeight(7)
      srtt:SetWidth(10)
      srtt:SetHeight(10)

      local srticon = ""
      srtt:SetText("")
      
      if p.GridParams.sortcol then
          for i, sortcol in ipairs(p.GridParams.sortcol) do
              if abs(sortcol) == col.pos then
                  srtt:SetText(i)
                  if sortcol < 0 then
                      srticon = "interface\\AddOns\\GridLib\\gfx\\arrow-u"
                  else
                      srticon = "interface\\AddOns\\GridLib\\gfx\\arrow-d"
                  end
              end
          end
      end
      srti:SetTexture(srticon)
      srti:SetTexCoord(0.25,0.75,0.125,0.875)
      srtt:Hide()
           
      
      if i == 1 then
       totalwidth = totalwidth  + (ColSepWidth/2);
      else
       totalwidth = totalwidth  + ColSepWidth;
      end
      ht:SetPoint("TOPLEFT", totalwidth , 0);
      ht:SetHeight(headerheight-1);
      ht:SetWidth(col.size);

      totalwidth = totalwidth  + col.size;
      
      local lbl = getglobal(ht:GetName().."Text");
      --tdbg(lbl:GetName())
      if lbl == nil then
       lbl = ht:CreateFontString(ht:GetName().."Text", "OVERLAY", "GameFontNormal");
      end
      if not h.FontObject then
       h.FontObject = CreateFont(p.gridid)
      end
      --local font, size, flags = lbl:GetFont();
      local font = "Fonts\\FRIZQT__.TTF"
      h.FontObject:SetFont(font, LabelFontSize)
      h.FontObject:SetShadowOffset(1,1)
      h.FontObject:SetShadowColor(0,0,0,0.7)
      h.FontObject:SetTextColor(unpack(HeaderTextColor))
      --ht:SetNormalFontObject(h.FontObject) 

      lbl:SetFontObject(h.FontObject) 

      local collbl = (col.lbl or "")

      if string.sub(collbl,1,2) == "#_" then
        lbl:SetText("");
      elseif string.sub(collbl,1,4) == "#GS_" then
        local foo = loadstring("return "..string.sub(collbl,5))
        lbl:SetText(foo());
      else
        lbl:SetText(MovixLib_GetLocale(collbl, p.gridid))
      end
      lbl:SetAllPoints();
      lbl:SetJustifyH(col.justify);
      lbl:SetJustifyV("MIDDLE");
      
      cvf = getglobal(h:GetName()..i.."VS");
      cvf:Show();
      cvf.i = i;
      cvf.hid = i;
      ht.hid = i;
      ht.parent = p;
      ht.p = p;
      if resetfilters then 
         ht.filtertbl = nil
      end
      --ht.col = col;      
      cvf.parent = p;
      cvf:SetPoint("TOPLEFT", totalwidth, 0);
      cvf:SetWidth(ColSepWidth);
      cvf:SetHeight(headerheight);
      
      cv = getglobal(h:GetName()..i.."VT");
      cv:Show();
      cv:SetPoint("TOPLEFT", (ColSepWidth-1)/2, 0);
      cv:SetWidth(1);
      cv:SetHeight(cvf:GetHeight());
      cv:SetTexture(unpack(ShadowColor2));
      cv = getglobal(h:GetName()..i.."VTS");
      cv:Show();
      cv:SetPoint("TOPLEFT", ((ColSepWidth-1)/2)+1, 0);
      cv:SetWidth(1);
      cv:SetHeight(cvf:GetHeight());
      cv:SetTexture(unpack(ShadowColor2));
      n = n + 1;
    end
   --end
   i = i + 1;
   local ht = getglobal(h:GetName()..i.."LBL");
   while (ht) do
     cvf = CreateFrame("FRAME", h:GetName()..i.."VS", scrollchild);
     cv = getglobal(h:GetName()..i.."VT");
     cv:Hide();
     cv = getglobal(h:GetName()..i.."VTS");
     cv:Hide();
     ht:Hide();
     cvf:Hide()
     i = i + 1
     ht = getglobal(h:GetName()..i.."LBL");
   end
  
   totalwidth = totalwidth + (ColSepWidth)
   
   scrollchild:SetWidth(totalwidth);
   local swidth = pwidth-(SliderSize+REdgeInset+LEdgeInset);
   if totalwidth < swidth then
    hslider:Hide();
    hslider:SetMinMaxValues(0,0); 
   else
    hslider:Show();
    hslider:SetMinMaxValues(0,totalwidth - swidth); 
   end
   p.columns = n;
end


function MovixLib_DrawGrid(p, rebuild)
 local t, l, w, h, totalw, LineHeight;

 local container = getglobal(p.HeaderLine:GetName().."Scrollchild");

 local gridid = p.gridid
 local LEdgeInset = Movixlib_GetGridParam(gridid, "LEdgeInset")
 local REdgeInset = Movixlib_GetGridParam(gridid, "REdgeInset") 
 local TEdgeInset = Movixlib_GetGridParam(gridid, "TEdgeInset")
 local BEdgeInset = Movixlib_GetGridParam(gridid, "BEdgeInset")
 local SliderSize = Movixlib_GetGridParam(gridid, "SliderSize")
 local headerheight = Movixlib_GetGridParam(gridid, "headerheight")
 local LineHeight = Movixlib_GetGridParam(gridid, "LineHeight")
 local ColSepWidth = Movixlib_GetGridParam(gridid, "ColSepWidth")
 local LabelFontSize = Movixlib_GetGridParam(gridid, "LabelFontSize")
 local GridBGcolor = Movixlib_GetGridParam(gridid, "GridBGcolor")
 local HeaderBGcolor = Movixlib_GetGridParam(gridid, "HeaderBGcolor")
 local OddLineColor = Movixlib_GetGridParam(gridid, "OddLineColor")
 local ShadowColor1 = Movixlib_GetGridParam(gridid, "ShadowColor1")
 local LightenColor1 = Movixlib_GetGridParam(gridid, "LightenColor1")
 local ShadowColor2 = Movixlib_GetGridParam(gridid, "ShadowColor2")
 local LightenColor2 = Movixlib_GetGridParam(gridid, "LightenColor2")
 local HeaderTextColor = Movixlib_GetGridParam(gridid, "HeaderTextColor")
   
 p.rebuild = rebuild;
 t = TEdgeInset+headerheight;
 l = LEdgeInset;
 w = p:GetWidth() - LEdgeInset - REdgeInset - SliderSize - 3
 h = p:GetHeight() - TEdgeInset - BEdgeInset - SliderSize -headerheight - 3
 p.dataGridHeight = h
 if rebuild then
  LineHeight = MovixLib_GetLineHeight(p)
  p.ComputedLineHeight = LineHeight 
  p.linecount = floor(h / LineHeight);
  if p.lines then 
    for k, gl in ipairs(p.lines) do 
--        tdbg("PARENT :", p:GetName(), " RELEASE LINE")
        if gl.cols then
            for ind, ctrl in ipairs(gl.cols) do
                local col = gl.cols[ind]
                col.ComputedW = nil
                col.ComputedH = nil
                col.Lpoint = nil
                col.Rpoint = nil
                col.root = nil
                col:Hide()
                gl.cols[ind] = nil
                MovixLib_ReleaseControlObject(ctrl:GetName())
            end
         else
         end
         gl.lineoptions = nil
         gl.parent = nil
         gl.cols = nil
         --gl.OddEven = nil
         gl:Hide()
         p.lines[k] = nil
    end
    p.lines= nil
  end
 end
 totalw = p:GetWidth()
 if p.lines == nil then
  --tdbg("PARENT :", p:GetName(), " LINES ARE NIL")
  p.lines = {}
 end
 local n = 0;
 local i
 local line = nil
 for i=1,p.linecount do
    local rebuildthisline = false
  
    if p.lines[i] == nil then
        --tdbg("PARENT :", p:GetName(), " LINE ", i, " is NIL")
        p.lines[i] = getglobal(p:GetName().."GridLine"..i)
        if not p.lines[i] then
            p.lines[i] = CreateFrame("BUTTON", p:GetName().."GridLine"..i, container) --, "MovixLib_lineTemplate");
        end
        rebuildthisline = true;
    end
    line = p.lines[i];
    line.lineoptions = MovixLib_NewLineOptions (p, i)
  
    line.parent = p;
    line:RegisterForClicks("AnyUp")
    line:SetScript("OnClick", MovixLib_OnClickLine)
    line:SetScript("OnDoubleClick", MovixLib_OnDoubleClickLine)
    line:SetScript("OnEnter", MovixLib_OnEnterLine)
    line:SetScript("OnLeave", MovixLib_OnLeaveLine)  
    line:ClearAllPoints()              
    --line:SetParent(container)
    line:SetHeight(LineHeight)
    if i == 1 then
        line:SetPoint("TOPLEFT", container, "TOPLEFT", -ColSepWidth/2, -t)--+1);
    else
        line:SetPoint("TOPLEFT", p.lines[i-1], "BOTTOMLEFT", 0, 0);
    end
  
    if fmod(i, 2) ~= 0 then
        if line.OddEven == nil then
            line.OddEven = line:CreateTexture(nil, "BACKGROUND")
        end
        line.OddEven:SetTexture(unpack(OddLineColor))
        line:SetNormalTexture(line.OddEven)
        line.OddEven:SetAllPoints()
    else
        line:SetNormalTexture("")
    end

    totalw = ColSepWidth/2;
  
    if line.cols == nil then
        line.cols = {} --MovixLib_GetTable();
    end
  
    if rebuildthisline then
        for ind, ctrl in ipairs(line.cols) do
            MovixLib_ReleaseControlObject(ctrl:GetName())
        end
    end

    local fld;
    local j = 0;
   local dbgcnt = 0
    for colid, col in ipairs(p.Headers) do
        local fld= nil;
        dbgcnt = dbgcnt + 1
        --if col.visible then
        j = j +1;
        if rebuildthisline or line.cols[j] == nil then
            --tdbg("PARENT :", p:GetName(), " LINE ", i, " COL ", j, " is NIL pass : ", dbgcnt)
            line.cols[j] = MovixLib_GetControl(line, col.type, col.pos, col.formfoo)
        end
    
        fld = line.cols[j] 
        fld.root = p;
        fld:SetParent(line)
        col.Lpoint = (ColSepWidth/2)+totalw
        fld:SetPoint("TOPLEFT", col.Lpoint,0);
        fld:SetHeight(LineHeight-2);
        fld:SetWidth(col.size);
        --col.Lpoint = col.Lpoint + (p.GridParams.ColSepWidth/2)
        col.Rpoint = totalw + col.size
        totalw = totalw + col.size + ColSepWidth;
        --end
    end
  
    j = j +1;
    while line.cols[j] do
        MovixLib_ReleaseControlObject(line.cols[j]:GetName()) 
        line.cols[j] =nil
        j = j +1;
    end
    --line:SetWidth(totalw)
    local lw = p:GetWidth()-SliderSize
    if lw < totalw then lw = totalw end
    line:SetWidth(lw)
    line:Hide()
  
    local hlt = getglobal(line:GetName().."HL")
    if not hlt then
        hlt = line:CreateTexture(line:GetName().."HL", "BACKGROUND");
    end
  
    hlt:SetTexture("interface\\targetingframe\\barfill2")
    hlt:SetTexCoord(0.2, 1, 0.6, 1)
    hlt:SetVertexColor(unpack(LightenColor1));
    hlt:SetAlpha(0.7)
    hlt:SetBlendMode("DISABLE")
    hlt:Hide()
    hlt:SetPoint("TOPLEFT",0,-1)
    hlt:SetPoint("BOTTOMRIGHT",0,-1)
    --hlt:SetWidth(lw)
    --hlt:SetHeight(LineHeight-2)
    n = i;
 end
 
 
 n = n +1;
 while p.lines[n] do
    for ind, ctrl in ipairs(p.lines[n].cols) do
        MovixLib_ReleaseControlObject(ctrl:GetName())
        p.lines[n].cols[ind] = nil
    end
    p.lines[n].cols = nil
    p.lines[n]:Hide();
    p.lines[n] = nil
    n = n +1;
 end

 p.mohlt_t = getglobal(p:GetName().."MOHL_T")
 if not p.mohlt_t then
   p.mohlt_t = p:CreateTexture(p:GetName().."MOHL_T", "OVERLAY");
   p.mohlt_b = p:CreateTexture(p:GetName().."MOHL_B", "OVERLAY");
   p.mohlt_t:SetTexture("interface\\targetingframe\\ui-targetingframe-barfill")
   p.mohlt_b:SetTexture("interface\\targetingframe\\ui-targetingframe-barfill")
   p.mohlt_t:SetTexCoord(0,1,0.6,0.9)
   p.mohlt_b:SetTexCoord(0,1,0.9,0.6)
   
   p.mohlt_t:SetVertexColor(1,1,1)
   p.mohlt_b:SetVertexColor(1,1,1)
   --p.mohlt_t:SetBlendMode("DISABLE")
   --p.mohlt_b:SetBlendMode("DISABLE")

   
   p.mohlt_t:SetPoint("TOPLEFT")
   p.mohlt_b:SetPoint("BOTTOMLEFT")
   p.mohlt_t:SetHeight(2)
   p.mohlt_b:SetHeight(2)
   
   
 end
 local lw = p:GetWidth()-SliderSize
 if lw < totalw then lw = totalw end
 p.mohlt_t:SetWidth(lw)
 p.mohlt_b:SetWidth(lw)

 p.mohlt_t:Hide()
 p.mohlt_b:Hide()

 
 
 local scrollit = getglobal(p:GetName().."_FauxScrollFrame");
 if scrollit == nil then
  scrollit = CreateFrame("SCROLLFRAME", p:GetName().."_FauxScrollFrame", p, "MovixLib_FauxScrollFrameTemplate");
 end
 --if rebuild then
    local scrollBar = getglobal(scrollit:GetName().."ScrollBar")
    local scrollbarslider = getglobal(scrollBar:GetName().."ThumbTexture")
    local scrollbaruparrow = getglobal(scrollBar:GetName().."ScrollUpButton")
    local scrollbardownarrow = getglobal(scrollBar:GetName().."ScrollDownButton")
    scrollbarslider:SetTexture("interface\\AddOns\\GridLib\\gfx\\bigslider")
    scrollbardownarrow:SetNormalTexture("interface\\AddOns\\GridLib\\gfx\\arrow-d")
    scrollbaruparrow:SetNormalTexture("interface\\AddOns\\GridLib\\gfx\\arrow-u")
    scrollbardownarrow:SetDisabledTexture("interface\\AddOns\\GridLib\\gfx\\arrow-d-d")
    scrollbaruparrow:SetDisabledTexture("interface\\AddOns\\GridLib\\gfx\\arrow-u-d")
    scrollbardownarrow:SetHighlightTexture("interface\\AddOns\\GridLib\\gfx\\arrow-d-h")
    scrollbaruparrow:SetHighlightTexture("interface\\AddOns\\GridLib\\gfx\\arrow-u-h")
    scrollbardownarrow:SetPushedTexture("interface\\AddOns\\GridLib\\gfx\\arrow-d-p")
    scrollbaruparrow:SetPushedTexture("interface\\AddOns\\GridLib\\gfx\\arrow-u-p")
    
    --scrollBar:SetValue(0);
 --end
 scrollit.parent=p;
 scrollit:SetPoint("TOPLEFT", p ,"TOPLEFT", l, -t)
 scrollit:SetHeight(h);
 scrollit:SetWidth(w);
 scrollbardownarrow:SetWidth(SliderSize);
 scrollbardownarrow:SetHeight(SliderSize);
 scrollbaruparrow:SetWidth(SliderSize);
 scrollbaruparrow:SetHeight(SliderSize);
 scrollbarslider:SetWidth(SliderSize-2);
 scrollbarslider:SetHeight(SliderSize-1);
 scrollbarslider:SetTexCoord(0.3,0.65,0.2,0.75)
 local sb = getglobal(p:GetName().."_FauxScrollFrameScrollBar");
 scrollBar:ClearAllPoints()
 scrollBar:SetPoint("TOPRIGHT", scrollit, "TOPRIGHT", SliderSize+2, -SliderSize-3);
 scrollBar:SetWidth(SliderSize-3)
 scrollBar:SetHeight(h-SliderSize-SliderSize-3);
 MovixLib_ScrollFrame_Update(scrollit);
end



